"""\
Search form extension for searching data indexed in Sphinx and PostgreSQL

Let's assume we want to use this as the *search* extension.

Install:

::

    wget pylonsbook.com/virtualenv.py
    python2.5 virtualenv.py ~/env
    ~/env/bin/easy_install SearchForm

Example Config:

::

    database.plugin = psycopg2
    database.pool = True
    database.database = carbage-dev
    database.user = carbage-dev
    database.password = carbage-dev
    database.host = localhost
    database.port = 5432

    search.template = main.dwt

    template.directory = ./

The template you specify should be a dreamweaver template relative to a Flows
application ``static/Templates`` directory or it should be an absolute path.
It should contain the following editable regions:

   ``doctitle``
       The page title in the ``<head>`` section. The ``<title>`` and
       ``</title>`` tags should be part of the region.

   ``head``
       An empty editable region in the ``<head>`` section which will be used
       to add CSS, JS and any required HTML tags.

   ``heading``
       The page heading text

   ``content``
       A region which will contain the form and any messages.

Example extension in ``ext/contact/dispatch.py``:

::

    import searchform
    
    on_load_component = searchform.make_component('contact')

That's it!

"""

import logging

from bn import AttributeDict, HTMLFragment, str_dict
from configconvert import handle_option_error, handle_section_error,\
   unicodeToObject
from flows.provider import start_flow, ServiceProvider
from conversionkit import Conversion, toDictionary, set_error, \
   chainConverters, chainPostConverters, toListOf
from formbuild import Form
from formconvert import multiDictToDict
from nestedrecord import decodeNestedRecord, encode_error
from recordconvert import toRecord, toListOfRecords
from stringconvert import unicodeToUnicode, unicodeToInteger
from httpkit.service.input import HTTPInputService
from dreamweavertemplate.service import DreamweaverTemplateService
from database.service.connection import DatabaseConnectionService
import sphinxapi

log = logging.getLogger(__name__)


search_form_converter = toRecord(
    filter_extra_fields = False,
    missing_or_empty_errors = {
        'q': 'Please enter some search terms',
    },
    converters = dict(
        q = unicodeToUnicode(),
    )
)

def config(flow, name):
    if not flow.config.option_group.has_key(name):
        flow.config.option_group['search'] = {}
        #raise handle_section_on_validation_error(
        #    flow, 
        #    name, 
        #    "'%s.template'"%(
        #        name
        #    )
        #)
    to_unicode = unicodeToUnicode()
    converter = chainConverters(
        decodeNestedRecord(),
        toRecord(
            missing_defaults = {
                'template': 'main.dwt',
            },
            empty_errors = {
                'template': 'Please enter a relative path from the Templates directory to the template filename you wish to use.',
            },
            converters = dict(
                template = to_unicode,
            ),
        ),
    )
    conversion = Conversion(flow.config.option_group[name]).perform(
        converter,
        flow,
    )
    if not conversion.successful:
        handle_option_on_validation_error(conversion, name)
    else:
        flow.config[name] = conversion.result
    return flow.config[name]

def on_generate_head_html(flow):
    return """\
<style type="text/css">
table.formbuild span.error-message, table.formbuild div.error, table.formbuild span.required {
    font-weight: bold;
    color: #f00;
}
table.formbuild span.small {
    font-size: 85%%;
}
table.formbuild form {
    margin-top: 20px;
}
table.formbuild form table td {
    padding-bottom: 3px;
}
table.formbuild label {
    font-weight: bold;
}
table.formbuild td.label {
    text-align: right;
}
</style>
"""

def on_valid_data(flow, result):   
    """\
    """
    terms = result['q']
    c = sphinxapi.SphinxClient()
    result = c.Query(terms)
    ids = [match['id'] for match in result['matches']]
    msg = 'No results'
    if ids:
        sql = 'SELECT * FROM search WHERE document_id in (%s)'
        strs = ('%s, '*len(ids))[:-2]
        sql = sql % (strs,)
        rows = flow.database.query(sql, ids)
        if not len(rows) == len(ids):
            log.error(
                '%s results from sphinx but %s from database', 
                len(ids),
                len(rows)
            )
        else:
            if len(ids) == 1:
                msg = '<p>1 Result...</p>'
            else:
                msg = '<p>%s Results...</p>'%(len(ids),)
            order = []
            cache = {}
            for row in rows:
                if row:
                    cache[row['document_id']] = row
            for result in result['matches']:
                order.append((result['weight'], result, cache[result['id']]))
            order.sort()
            order.reverse()
            fragment = HTMLFragment()
            fragment.safe("""     <table border="0"> """)
            for weight, result, cache in order:
                fragment.safe('''     <tr><td><b><a href="''')
                fragment.write(cache['path'])
                fragment.safe('''">''')
                fragment.write(cache['title'])
                fragment.safe("""     </a></b></td><td>Weight: """)
                fragment.write(str(weight))
                fragment.safe("""     </td></tr><tr><td colspan="2"> """)
                fragment.write(cache['description'])
                fragment.safe("""     </td></tr><tr><td colspan="2">&nbsp; <br /><td></tr> """)
            fragment.safe("""     </table> """)
            msg += fragment.getvalue()
    return flow.template.render(
        flow.config.ext.template, 
        {
            'doctitle': '<title>Search Form</title>',
            'heading': 'Search Form',
            'content': msg,
            'head': on_generate_head_html(flow),
        } 
    )

def on_generate_form_html(flow):
    form = flow.ext.form
    fragment = HTMLFragment()
    fragment.safe(form.start_with_layout(''))
    fragment.safe(
        form.field(
            label='Search',
            type='text', 
            name='q', 
            attributes=dict(size=38),
            required=True,
        )
    )
    actions = form.submit(name='submit', value='Search')
    fragment.safe(form.action_bar(actions))
    fragment.safe(form.end_with_layout())
    return fragment.getvalue()
    
def on_new_form(flow):
    return flow.template.render(
        flow.config.ext.template, 
        {
            'doctitle': '<title>Search Form</title>',
            'heading': 'Search Form',
            'content': (
                '<p>Enter search terms:</p>'+\
                flow.ext.on_generate_form_html(flow)
            ),
            'head': flow.ext.on_generate_head_html(flow),
        }
    )

def on_validation_error(flow, form):
    """\
    """
    # We happen to be able to use the same function:
    return on_new_form(flow)

def make_component(
    name,
    on_new_form = on_new_form,
    #on_validation_error = on_validation_error,
    #on_valid_data = on_valid_data,
    on_generate_form_html = on_generate_form_html,
    on_generate_head_html = on_generate_head_html,
):

    def run(flow):
        flow['ext'] = AttributeDict(
            on_generate_form_html=on_generate_form_html,
            on_generate_head_html=on_generate_head_html,
        )
        if not flow.http.input:
            form = Form(
                value=dict(
                ),
            )
            flow.ext['form'] = form
            flow.http.response.body = [on_new_form(flow)]
        else:
            # Handle the submission
            conversion = Conversion(
                flow.http.input
            ).perform(
                multiDictToDict(encoding='utf8'),
                flow,
            )
            if not conversion.successful:
                raise Exception(conversion.error)
            params = conversion.result
            conversion = Conversion(
                params
            ).perform(
                search_form_converter,
                flow,
            )
            if not conversion.successful:
                form = Form(
                    value=conversion.value, 
                    error=encode_error(conversion),
                )
                flow.ext['form'] = form
                flow.http.response.body = [on_validation_error(flow, form)]
            else:
                flow.http.response.body = [on_valid_data(
                    flow,
                    conversion.result,
                )]

    required_services = ['http.input', 'template']
    available_services = {
        'http.input': HTTPInputService,
        'database': DatabaseConnectionService,
        'template': DreamweaverTemplateService,
    }

    def on_load_component(flow):
        flow.config[name] = config(flow, name)
        #if flow.config.has_key('ext'):
        #    raise Exception('Extension configuration is already present')
        #else:
        flow.config['ext'] = flow.config[name]
        def dispatch_contact_form(flow):
            f = flow.copy()
            del f['provider']
            start_flow(
                run,
                provider = ServiceProvider(
                    existing_services = f,
                    available_services = available_services,
                ),
                required_services=required_services,
            )
        return dispatch_contact_form
    return on_load_component

